@@ -336,6 +336,9 @@ func archPPC64(linkArch *obj.LinkArch) *Arch {
 	for i := ppc64.REG_VS0; i <= ppc64.REG_VS63; i++ {
 		register[obj.Rconv(i)] = int16(i)
 	}
+	for i := ppc64.REG_A0; i <= ppc64.REG_A7; i++ {
+		register[obj.Rconv(i)] = int16(i)
+	}
 	for i := ppc64.REG_CR0; i <= ppc64.REG_CR7; i++ {
 		register[obj.Rconv(i)] = int16(i)
 	}
@@ -374,6 +377,11 @@ func archPPC64(linkArch *obj.LinkArch) *Arch {
 			instructions[s] = obj.As(i) + obj.ABasePPC64
 		}
 	}
+	// The opcodes generated by x/arch's ppc64map are listed in
+	// a separate slice, add them too.
+	for i, s := range ppc64.GenAnames {
+		instructions[s] = obj.As(i) + ppc64.AFIRSTGEN
+	}
 	// Annoying aliases.
 	instructions["BR"] = ppc64.ABR
 	instructions["BL"] = ppc64.ABL
@@ -528,7 +536,6 @@ func archLoong64(linkArch *obj.LinkArch) *Arch {
 	// Avoid unintentionally clobbering g using R22.
 	delete(register, "R22")
 	register["g"] = loong64.REG_R22
-	register["RSB"] = loong64.REG_R31
 	registerPrefix := map[string]bool{
 		"F":    true,
 		"FCSR": true,
