@@ -98,8 +98,9 @@ var (
 	flagDebugNosplit  = flag.Bool("debugnosplit", false, "dump nosplit call graph")
 	FlagStrictDups    = flag.Int("strictdups", 0, "sanity check duplicate symbol contents during object file reading (1=warn 2=err).")
 	FlagRound         = flag.Int("R", -1, "set address rounding `quantum`")
-	FlagTextAddr      = flag.Int64("T", -1, "set text segment `address`")
+	FlagTextAddr      = flag.Int64("T", -1, "set the start address of text symbols")
 	flagEntrySymbol   = flag.String("E", "", "set `entry` symbol name")
+	flagPruneWeakMap  = flag.Bool("pruneweakmap", true, "prune weak mapinit refs")
 	cpuprofile        = flag.String("cpuprofile", "", "write cpu profile to `file`")
 	memprofile        = flag.String("memprofile", "", "write memory profile to `file`")
 	memprofilerate    = flag.Int64("memprofilerate", 0, "set runtime.MemProfileRate to `rate`")
@@ -109,6 +110,9 @@ var (
 
 // Main is the main entry point for the linker code.
 func Main(arch *sys.Arch, theArch Arch) {
+	log.SetPrefix("link: ")
+	log.SetFlags(0)
+
 	thearch = theArch
 	ctxt := linknew(arch)
 	ctxt.Bso = bufio.NewWriter(os.Stdout)
@@ -158,6 +162,14 @@ func Main(arch *sys.Arch, theArch Arch) {
 		// dump symbol info on crash
 		defer func() { ctxt.loader.Dump() }()
 	}
+	if ctxt.Debugvlog > 1 {
+		// dump symbol info on error
+		AtExit(func() {
+			if nerrors > 0 {
+				ctxt.loader.Dump()
+			}
+		})
+	}
 
 	switch *flagHeadType {
 	case "":
@@ -267,6 +279,9 @@ func Main(arch *sys.Arch, theArch Arch) {
 	bench.Start("loadlib")
 	ctxt.loadlib()
 
+	bench.Start("inittasks")
+	ctxt.inittasks()
+
 	bench.Start("deadcode")
 	deadcode(ctxt)
 
@@ -343,12 +358,6 @@ func Main(arch *sys.Arch, theArch Arch) {
 	bench.Start("layout")
 	filesize := ctxt.layout(order)
 
-	// Fix up symbols just prior to output.
-	bench.Start("FixUpSyms")
-	if thearch.FixUpSyms != nil {
-		thearch.FixUpSyms(ctxt, ctxt.loader)
-	}
-
 	// Write out the output file.
 	// It is split into two parts (Asmb and Asmb2). The first
 	// part writes most of the content (sections and segments),
