@@ -72,12 +72,18 @@ func gentext(ctxt *ld.Link, ldr *loader.Loader) {
 	ld.Elfwritedynent(ctxt.Arch, dynamic, elf.DT_MIPS_RLD_VERSION, 1)
 	ld.Elfwritedynent(ctxt.Arch, dynamic, elf.DT_MIPS_BASE_ADDRESS, 0)
 
+	// elfsetupplt should have been called and gotLocalCount should now
+	// have its correct value.
+	if gotLocalCount == 0 {
+		ctxt.Errorf(0, "internal error: elfsetupplt has not been called")
+	}
+	ld.Elfwritedynent(ctxt.Arch, dynamic, elf.DT_MIPS_LOCAL_GOTNO, gotLocalCount)
+
 	// DT_* entries have to exist prior to elfdynhash(), which finalises the
 	// table by adding DT_NULL. However, the values for the following entries
 	// are not know until after dynreloc() has completed. Add the symbols now,
 	// then update their values prior to code generation.
 	dts := []elf.DynTag{
-		elf.DT_MIPS_LOCAL_GOTNO,
 		elf.DT_MIPS_SYMTABNO,
 		elf.DT_MIPS_GOTSYM,
 	}
@@ -88,18 +94,6 @@ func gentext(ctxt *ld.Link, ldr *loader.Loader) {
 	}
 }
 
-func fixUpSyms(ctxt *ld.Link, ldr *loader.Loader) {
-	if *ld.FlagD || ctxt.Target.IsExternal() {
-		return
-	}
-
-	// Update DT_* entries now that we have the values to do so.
-	dynamic := ldr.MakeSymbolUpdater(ctxt.ArchSyms.Dynamic)
-	dynamic.SetUint(ctxt.Target.Arch, dtOffsets[elf.DT_MIPS_LOCAL_GOTNO], gotLocalCount)
-	dynamic.SetUint(ctxt.Target.Arch, dtOffsets[elf.DT_MIPS_SYMTABNO], dynSymCount)
-	dynamic.SetUint(ctxt.Target.Arch, dtOffsets[elf.DT_MIPS_GOTSYM], gotSymIndex)
-}
-
 func adddynrel(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loader.Sym, r loader.Reloc, rIdx int) bool {
 	targ := r.Sym()
 	var targType sym.SymKind
@@ -142,51 +136,41 @@ func elfreloc1(ctxt *ld.Link, out *ld.OutBuf, ldr *loader.Loader, s loader.Sym,
 
 	// mips64 ELF relocation (endian neutral)
 	//		offset	uint64
-	//		sym		uint32
+	//		sym	uint32
 	//		ssym	uint8
 	//		type3	uint8
 	//		type2	uint8
-	//		type	uint8
+	//		type1	uint8
 	//		addend	int64
 
 	addend := r.Xadd
 
+	// offset
 	out.Write64(uint64(sectoff))
 
+	// sym and ssym
 	elfsym := ld.ElfSymForReloc(ctxt, r.Xsym)
 	out.Write32(uint32(elfsym))
 	out.Write8(0)
 
+	// types
+	var type3, type2, type1 elf.R_MIPS
 	switch r.Type {
-	case objabi.R_MIPS_GPREL_HI16:
-		out.Write8(uint8(elf.R_MIPS_HI16))
-		out.Write8(uint8(elf.R_MIPS_SUB))
-	case objabi.R_MIPS_GPREL_LO16:
-		out.Write8(uint8(elf.R_MIPS_LO16))
-		out.Write8(uint8(elf.R_MIPS_SUB))
-	default:
-		out.Write8(0)
-		out.Write8(0)
-	}
-
-	switch r.Type {
-	default:
-		return false
 	case objabi.R_ADDR, objabi.R_DWARFSECREF:
 		switch r.Size {
 		case 4:
-			out.Write8(uint8(elf.R_MIPS_32))
+			type1 = elf.R_MIPS_32
 		case 8:
-			out.Write8(uint8(elf.R_MIPS_64))
+			type1 = elf.R_MIPS_64
 		default:
 			return false
 		}
 	case objabi.R_ADDRMIPS:
-		out.Write8(uint8(elf.R_MIPS_LO16))
+		type1 = elf.R_MIPS_LO16
 	case objabi.R_ADDRMIPSU:
-		out.Write8(uint8(elf.R_MIPS_HI16))
+		type1 = elf.R_MIPS_HI16
 	case objabi.R_ADDRMIPSTLS:
-		out.Write8(uint8(elf.R_MIPS_TLS_TPREL_LO16))
+		type1 = elf.R_MIPS_TLS_TPREL_LO16
 		if ctxt.Target.IsOpenbsd() {
 			// OpenBSD mips64 does not currently offset TLS by 0x7000,
 			// as such we need to add this back to get the correct offset
@@ -194,16 +178,23 @@ func elfreloc1(ctxt *ld.Link, out *ld.OutBuf, ldr *loader.Loader, s loader.Sym,
 			addend += 0x7000
 		}
 	case objabi.R_CALLMIPS, objabi.R_JMPMIPS:
-		out.Write8(uint8(elf.R_MIPS_26))
+		type1 = elf.R_MIPS_26
 	case objabi.R_MIPS_GPREL_HI16:
-		out.Write8(uint8(elf.R_MIPS_GPREL16))
+		type3, type2, type1 = elf.R_MIPS_HI16, elf.R_MIPS_SUB, elf.R_MIPS_GPREL16
 	case objabi.R_MIPS_GPREL_LO16:
-		out.Write8(uint8(elf.R_MIPS_GPREL16))
+		type3, type2, type1 = elf.R_MIPS_LO16, elf.R_MIPS_SUB, elf.R_MIPS_GPREL16
 	case objabi.R_MIPS_CALL16:
-		out.Write8(uint8(elf.R_MIPS_CALL16))
+		type1 = elf.R_MIPS_CALL16
 	case objabi.R_MIPS_JALR:
-		out.Write8(uint8(elf.R_MIPS_JALR))
+		type1 = elf.R_MIPS_JALR
+	default:
+		return false
 	}
+	out.Write8(uint8(type3))
+	out.Write8(uint8(type2))
+	out.Write8(uint8(type1))
+
+	// addend
 	out.Write64(uint64(addend))
 
 	return true
@@ -251,10 +242,18 @@ func addpltsym(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 		return
 	}
 
+	dynamic := ldr.MakeSymbolUpdater(syms.Dynamic)
+
+	const dynSymEntrySize = 20
 	if gotSymIndex == 0 {
-		const dynSymEntrySize = 20
+		// Compute and update GOT symbol index.
 		gotSymIndex = uint64(ldr.SymSize(syms.DynSym) / dynSymEntrySize)
+		dynamic.SetUint(target.Arch, dtOffsets[elf.DT_MIPS_GOTSYM], gotSymIndex)
 	}
+	if dynSymCount == 0 {
+		dynSymCount = uint64(ldr.SymSize(syms.DynSym) / dynSymEntrySize)
+	}
+
 	ld.Adddynsym(ldr, target, syms, s)
 	dynSymCount++
 
@@ -283,6 +282,9 @@ func addpltsym(target *ld.Target, ldr *loader.Loader, syms *ld.ArchSyms, s loade
 	gotplt.AddAddrPlus(target.Arch, plt.Sym(), 0)
 
 	ldr.SetPlt(s, int32(plt.Size()-16))
+
+	// Update dynamic symbol count.
+	dynamic.SetUint(target.Arch, dtOffsets[elf.DT_MIPS_SYMTABNO], dynSymCount)
 }
 
 func machoreloc1(*sys.Arch, *ld.OutBuf, *loader.Loader, loader.Sym, loader.ExtReloc, int64) bool {
