@@ -25,9 +25,6 @@ func Init() (*sys.Arch, ld.Arch) {
 		Archreloc:        archreloc,
 		Archrelocvariant: archrelocvariant,
 		Extreloc:         extreloc,
-		Elfreloc1:        elfreloc1,
-		ElfrelocSize:     24,
-		Elfsetupplt:      elfsetupplt,
 
 		// TrampLimit is set such that we always run the trampoline
 		// generation code. This is necessary since calls to external
@@ -40,12 +37,19 @@ func Init() (*sys.Arch, ld.Arch) {
 		GenSymsLate: genSymsLate,
 		Machoreloc1: machoreloc1,
 
-		Dragonflydynld: "XXX",
-		Freebsddynld:   "/usr/libexec/ld-elf.so.1",
-		Linuxdynld:     "/lib/ld.so.1",
-		Netbsddynld:    "XXX",
-		Openbsddynld:   "/usr/libexec/ld.so",
-		Solarisdynld:   "XXX",
+		ELF: ld.ELFArch{
+			Linuxdynld: "/lib/ld.so.1",
+
+			Freebsddynld:   "/usr/libexec/ld-elf.so.1",
+			Netbsddynld:    "XXX",
+			Openbsddynld:   "/usr/libexec/ld.so",
+			Dragonflydynld: "XXX",
+			Solarisdynld:   "XXX",
+
+			Reloc1:    elfreloc1,
+			RelocSize: 24,
+			SetupPLT:  elfsetupplt,
+		},
 	}
 
 	return arch, theArch
