@@ -6,7 +6,10 @@
 
 package runtime
 
-import "unsafe"
+import (
+	"internal/abi"
+	"unsafe"
+)
 
 const (
 	debugCallSystemStack = "executing on Go runtime stack"
@@ -80,8 +83,8 @@ func debugCallCheck(pc uintptr) string {
 		if pc != f.entry() {
 			pc--
 		}
-		up := pcdatavalue(f, _PCDATA_UnsafePoint, pc, nil)
-		if up != _PCDATA_UnsafePointSafe {
+		up := pcdatavalue(f, abi.PCDATA_UnsafePoint, pc, nil)
+		if up != abi.UnsafePointSafe {
 			// Not at a safe point.
 			ret = debugCallUnsafePoint
 		}
@@ -158,8 +161,8 @@ func debugCallWrap(dispatch uintptr) {
 		gp.schedlink = 0
 
 		// Park the calling goroutine.
-		if trace.enabled {
-			traceGoPark(traceEvGoBlock, 1)
+		if traceEnabled() {
+			traceGoPark(traceBlockDebugCall, 1)
 		}
 		casGToWaiting(gp, _Grunning, waitReasonDebugCall)
 		dropg()
@@ -217,7 +220,7 @@ func debugCallWrap1() {
 		// Switch back to the calling goroutine. At some point
 		// the scheduler will schedule us again and we'll
 		// finish exiting.
-		if trace.enabled {
+		if traceEnabled() {
 			traceGoSched()
 		}
 		casgstatus(gp, _Grunning, _Grunnable)
@@ -226,7 +229,7 @@ func debugCallWrap1() {
 		globrunqput(gp)
 		unlock(&sched.lock)
 
-		if trace.enabled {
+		if traceEnabled() {
 			traceGoUnpark(callingG, 0)
 		}
 		casgstatus(callingG, _Gwaiting, _Grunnable)
