@@ -6,14 +6,9 @@
 
 package runtime
 
-import (
-	"runtime/internal/syscall"
-)
-
 const SiginfoMaxSize = _si_max_size
 const SigeventMaxSize = _sigev_max_size
 
-var Closeonexec = syscall.CloseOnExec
 var NewOSProc0 = newosproc0
 var Mincore = mincore
 var Add = add
