@@ -9,3 +9,7 @@ package runtime
 // Called from assembly only; declared for go vet.
 func load_g()
 func save_g()
+
+// getfp returns the frame pointer register of its caller or 0 if not implemented.
+// TODO: Make this a compiler intrinsic
+func getfp() uintptr { return 0 }
